"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _lokijs = _interopRequireDefault(require("lokijs"));

var _underscore = _interopRequireDefault(require("underscore"));

var _enum = require("./enum");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var truePredicate = function truePredicate() {
  return true;
};

var finishedPredicate = function finishedPredicate(req) {
  return req.success || req.error || req.timeout;
};

var averageCalc = function averageCalc(a, n, b, m) {
  return a * (n / (n + m)) + b * (m / (n + m));
};

var MetricUtils = {
  createRequest: function createRequest(envelop) {
    return {
      id: envelop.id,
      event: envelop.tag,
      from: envelop.owner,
      to: envelop.recipient,
      size: [envelop.size],
      timeout: false,
      duration: {
        latency: -1,
        process: -1
      },
      success: false,
      error: false
    };
  },
  createTick: function createTick(envelop) {
    return {
      id: envelop.id,
      event: envelop.tag,
      from: envelop.owner,
      to: envelop.recipient,
      size: envelop.size
    };
  }
};

var _private = new WeakMap();

var _updateAggregationTable = function _updateAggregationTable() {
  var _scope = _private.get(this);

  var aggregationTable = _scope.aggregationTable,
      customColumns = _scope.customColumns; // resetting timeout and count

  _scope.count = 0;
  clearTimeout(_scope.flushTimeoutInstance);
  _scope.flushTimeoutInstance = setTimeout(_updateAggregationTable.bind(this), _scope.flushTimeout); // getting requests and ticks

  var sendRequests = _scope.sendRequestCollection.where(finishedPredicate);

  var gotRequests = _scope.gotRequestCollection.where(finishedPredicate);

  var sendTicks = _scope.sendTickCollection.where(truePredicate);

  var gotTicks = _scope.gotTickCollection.where(truePredicate); // grouping by node and event


  sendRequests = _underscore["default"].groupBy(sendRequests, function (request) {
    return "".concat(request.to).concat(request.event);
  });
  gotRequests = _underscore["default"].groupBy(gotRequests, function (request) {
    return "".concat(request.from).concat(request.event);
  });
  sendTicks = _underscore["default"].groupBy(sendTicks, function (tick) {
    return "".concat(tick.to).concat(tick.event);
  });
  gotTicks = _underscore["default"].groupBy(gotTicks, function (tick) {
    return "".concat(tick.from).concat(tick.event);
  }); // updating row in aggregation table

  var updateRequestRow = function updateRequestRow(node, event, groupedRequests) {
    var out = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;
    var row = aggregationTable.findOne({
      node: node,
      event: event,
      out: out,
      request: true
    });

    if (!row) {
      row = {
        node: node,
        event: event,
        out: out,
        request: true,
        latency: 0,
        process: 0,
        count: 0,
        success: 0,
        error: 0,
        timeout: 0,
        size: 0
      };

      _underscore["default"].each(customColumns, function (_ref, columnName) {
        var initialValue = _ref.initialValue;
        row[columnName] = initialValue;
      });

      row = aggregationTable.insert(row);
    }

    var latencySum = 0;
    var processSum = 0;
    var sizeSum = 0;
    var allCount = row.count;
    var initialCount = row.count - row.timeout;

    _underscore["default"].each(groupedRequests, function (request) {
      row.count++;
      row.success += request.success;
      row.error += request.error;
      row.timeout += request.timeout;
      latencySum += request.duration.latency;
      processSum += request.duration.process;
      sizeSum += request.size[0] + request.size[1];

      _underscore["default"].each(customColumns, function (_ref2, columnName) {
        var reducer = _ref2.reducer;
        row[columnName] = reducer(row, request);
      });
    });

    row.latency = averageCalc(row.latency, initialCount, latencySum / (row.count - initialCount - row.timeout), row.count - initialCount - row.timeout);
    row.process = averageCalc(row.process, initialCount, processSum / (row.count - initialCount - row.timeout), row.count - initialCount - row.timeout);
    row.size = averageCalc(row.size, allCount, sizeSum / (row.count - allCount), row.count - allCount);
    aggregationTable.update(row);
  }; // updating row in aggregation table


  var updateTickRow = function updateTickRow(node, event, groupedTicks) {
    var out = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;
    var row = aggregationTable.find({
      node: node,
      event: event,
      out: out,
      request: false
    });

    if (!row) {
      row = {
        node: node,
        event: event,
        out: out,
        request: false,
        count: 0,
        size: 0
      };

      _underscore["default"].each(customColumns, function (_ref3, columnName) {
        var initialValue = _ref3.initialValue;
        row[columnName] = initialValue;
      });

      row = aggregationTable.insert(row);
    }

    var sizeSum = 0;
    var initialCount = row.count;

    _underscore["default"].each(groupedTicks, function (request) {
      row.count++;
      sizeSum += request.size;

      _underscore["default"].each(customColumns, function (_ref4, columnName) {
        var reducer = _ref4.reducer;
        row[columnName] = reducer(row, request);
      });
    });

    row.size = averageCalc(row.size, initialCount, sizeSum / groupedTicks.length, groupedTicks.length);
    aggregationTable.update(row);
  };

  _underscore["default"].each(sendRequests, function (groupedRequests) {
    updateRequestRow(groupedRequests[0].to, groupedRequests[0].event, groupedRequests, true);
  });

  _underscore["default"].each(gotRequests, function (groupedRequests) {
    updateRequestRow(groupedRequests[0].from, groupedRequests[0].event, groupedRequests);
  });

  _underscore["default"].each(sendTicks, function (groupedTicks) {
    updateTickRow(groupedTicks[0].from, groupedTicks[0].event, groupedTicks, true);
  });

  _underscore["default"].each(gotTicks, function (groupedTicks) {
    updateTickRow(groupedTicks[0].from, groupedTicks[0].event, groupedTicks);
  });

  this.flush();
};

var Metric =
/*#__PURE__*/
function () {
  function Metric() {
    var _ref5 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
        id = _ref5.id;

    _classCallCheck(this, Metric);

    var ZeronodeMetricDB = new _lokijs["default"]('zeronode.db');
    var _scope = {
      id: id,
      enabled: false,
      // ** loki collections
      sendRequestCollection: ZeronodeMetricDB.addCollection(_enum.MetricCollections.SEND_REQUEST, {
        indices: ['id']
      }),
      gotRequestCollection: ZeronodeMetricDB.addCollection(_enum.MetricCollections.GOT_REQUEST, {
        indices: ['id']
      }),
      sendTickCollection: ZeronodeMetricDB.addCollection(_enum.MetricCollections.SEND_TICK, {
        indices: ['id']
      }),
      gotTickCollection: ZeronodeMetricDB.addCollection(_enum.MetricCollections.GOT_TICK, {
        indices: ['id']
      }),
      aggregationTable: ZeronodeMetricDB.addCollection(_enum.MetricCollections.AGGREGATION, {
        indices: ['node', 'event']
      }),
      flushTimeoutInstance: null,
      flushTimeout: 30 * 1000,
      customColumns: {},
      count: 0
    };

    _private.set(this, _scope);

    this.db = ZeronodeMetricDB;
  }

  _createClass(Metric, [{
    key: "getMetrics",
    value: function getMetrics() {
      var query = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

      var _private$get = _private.get(this),
          aggregationTable = _private$get.aggregationTable;

      if (!this.status) return;
      var result = aggregationTable.find(query);
      var total = {
        count: 0,
        latency: 0,
        process: 0,
        out: 0,
        "in": 0,
        request: 0,
        tick: 0,
        error: 0,
        success: 0,
        timeout: 0,
        size: 0 // calculating total

      };
      total = _underscore["default"].reduce(result, function (memo, row) {
        var initialCount = memo.count;
        var initialOut = memo.out;
        var initialTimeout = memo.timeout;
        memo.count += row.count;
        row.out ? memo.out += row.count : memo["in"] += row.count;
        row.request ? memo.request += row.count : memo.tick += row.count;

        if (row.request) {
          memo.error += row.error;
          memo.success += row.success;
          memo.timeout += row.timeout;

          if (row.out) {
            memo.latency = averageCalc(memo.latency, initialOut - initialTimeout, row.latency, row.count - row.timeout);
            memo.process = averageCalc(memo.process, initialOut - initialTimeout, row.process, row.count - row.timeout);
          }
        }

        memo.size = averageCalc(memo.size, initialCount, row.size, row.count);
        return memo;
      }, total);
      return {
        result: result,
        total: total
      };
    }
  }, {
    key: "defineColumn",
    value: function defineColumn(columnName, initialValue, reducer) {
      var isIndex = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;

      var _private$get2 = _private.get(this),
          aggregationTable = _private$get2.aggregationTable,
          customColumns = _private$get2.customColumns;

      if (this.status) throw new Error("Can't define column after metrics enabled");

      if (isIndex) {
        aggregationTable.ensureIndex(columnName);
      }

      customColumns[columnName] = {
        initialValue: initialValue,
        reducer: reducer,
        isIndex: isIndex
      };
    } //TODO:: avar, dave

  }, {
    key: "enable",
    value: function enable(flushTimeout) {
      var _scope = _private.get(this);

      _scope.enabled = true;
      _scope.flushTimeout = flushTimeout || _scope.flushTimeout;
      _scope.flushTimeoutInstance = setTimeout(_updateAggregationTable.bind(this), _scope.flushTimeout);
    } //TODO:: avar, dave

  }, {
    key: "disable",
    value: function disable() {
      var _scope = _private.get(this);

      _scope.enabled = false;
      clearTimeout(_scope.flushTimeoutInstance);

      _updateAggregationTable.call(this);

      _scope.count = 0;
    } // ** actions

  }, {
    key: "sendRequest",
    value: function sendRequest(envelop) {
      var _private$get3 = _private.get(this),
          sendRequestCollection = _private$get3.sendRequestCollection,
          enabled = _private$get3.enabled;

      if (!enabled) return;
      var requestInstance = MetricUtils.createRequest(envelop);
      sendRequestCollection.insert(requestInstance);
    }
  }, {
    key: "gotRequest",
    value: function gotRequest(envelop) {
      var _private$get4 = _private.get(this),
          gotRequestCollection = _private$get4.gotRequestCollection,
          enabled = _private$get4.enabled;

      if (!enabled) return;
      var requestInstance = MetricUtils.createRequest(envelop);
      gotRequestCollection.insert(requestInstance);
    }
  }, {
    key: "sendReplySuccess",
    value: function sendReplySuccess(envelop) {
      var _scope = _private.get(this);

      var gotRequestCollection = _scope.gotRequestCollection,
          enabled = _scope.enabled;
      if (!enabled) return;
      var request = gotRequestCollection.findOne({
        id: envelop.id
      });
      if (!request) return;
      request.success = true;
      request.size.push(envelop.size);
      gotRequestCollection.update(request);

      if (++_scope.count === 1000) {
        _updateAggregationTable.call(this);
      }
    }
  }, {
    key: "sendReplyError",
    value: function sendReplyError(envelop) {
      var _scope = _private.get(this);

      var gotRequestCollection = _scope.gotRequestCollection,
          enabled = _scope.enabled;
      if (!enabled) return;
      var request = gotRequestCollection.findOne({
        id: envelop.id
      });
      if (!request) return;
      request.error = true;
      request.size.push(envelop.size);
      gotRequestCollection.update(request);

      if (++_scope.count === 1000) {
        _updateAggregationTable.call(this);
      }
    }
  }, {
    key: "gotReplySuccess",
    value: function gotReplySuccess(envelop) {
      var _scope = _private.get(this);

      var sendRequestCollection = _scope.sendRequestCollection,
          enabled = _scope.enabled;
      if (!enabled) return;
      var request = sendRequestCollection.findOne({
        id: envelop.id
      });
      if (!request) return;
      request.success = true;
      request.duration = envelop.data.duration;
      request.size.push(envelop.size);
      sendRequestCollection.update(request);

      if (++_scope.count === 1000) {
        _updateAggregationTable.call(this);
      }
    }
  }, {
    key: "gotReplyError",
    value: function gotReplyError(envelop) {
      var _scope = _private.get(this);

      var sendRequestCollection = _scope.sendRequestCollection,
          enabled = _scope.enabled;
      if (!enabled) return;
      var request = sendRequestCollection.findOne({
        id: envelop.id
      });
      if (!request) return;
      request.error = true;
      request.duration = envelop.data.duration;
      request.size.push(envelop.size);
      sendRequestCollection.update(request);
    }
  }, {
    key: "requestTimeout",
    value: function requestTimeout(envelop) {
      var _scope = _private.get(this);

      var sendRequestCollection = _scope.sendRequestCollection,
          enabled = _scope.enabled;
      if (!enabled) return;
      var request = sendRequestCollection.findOne({
        id: envelop.id
      });
      if (!request) return;
      request.timeout = true;
      sendRequestCollection.update(request);

      if (++_scope.count === 1000) {
        _updateAggregationTable.call(this);
      }
    }
  }, {
    key: "sendTick",
    value: function sendTick(envelop) {
      var _scope = _private.get(this);

      var sendTickCollection = _scope.sendTickCollection,
          enabled = _scope.enabled;
      if (!enabled) return;
      var tickInstance = MetricUtils.createTick(envelop);
      sendTickCollection.insert(tickInstance);

      if (++_scope.count === 1000) {
        _updateAggregationTable.call(this);
      }
    }
  }, {
    key: "gotTick",
    value: function gotTick(envelop) {
      var _scope = _private.get(this);

      var gotTickCollection = _scope.gotTickCollection,
          enabled = _scope.enabled;
      if (!enabled) return;
      var tickInstance = MetricUtils.createTick(envelop);
      gotTickCollection.insert(tickInstance);

      if (++_scope.count === 1000) {
        _updateAggregationTable.call(this);
      }
    }
  }, {
    key: "flush",
    value: function flush() {
      var _scope = _private.get(this);

      var sendRequestCollection = _scope.sendRequestCollection,
          sendTickCollection = _scope.sendTickCollection,
          gotRequestCollection = _scope.gotRequestCollection,
          gotTickCollection = _scope.gotTickCollection;
      sendRequestCollection.removeWhere(finishedPredicate);
      sendTickCollection.removeWhere(finishedPredicate);
      gotRequestCollection.removeWhere(truePredicate);
      gotTickCollection.removeWhere(truePredicate);
      _scope.count = 0;
    }
  }, {
    key: "status",
    get: function get() {
      var _private$get5 = _private.get(this),
          enabled = _private$get5.enabled;

      return enabled;
    }
  }]);

  return Metric;
}();

exports["default"] = Metric;