"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _winston = _interopRequireDefault(require("winston"));

var _underscore = _interopRequireDefault(require("underscore"));

var _bluebird = _interopRequireDefault(require("bluebird"));

var _md = _interopRequireDefault(require("md5"));

var _animalId = _interopRequireDefault(require("animal-id"));

var _events = require("events");

var _errors = require("./errors");

var _utils = _interopRequireDefault(require("./utils"));

var _server = _interopRequireDefault(require("./server"));

var _client2 = _interopRequireDefault(require("./client"));

var _metric = _interopRequireDefault(require("./metric"));

var _enum = require("./enum");

var _sockets = require("./sockets");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var MetricType = _sockets.Enum.MetricType;

var _private = new WeakMap();

var defaultLogger = _winston["default"].createLogger({
  transports: [new _winston["default"].transports.Console({
    level: 'error'
  })]
});

var Node =
/*#__PURE__*/
function (_EventEmitter) {
  _inherits(Node, _EventEmitter);

  function Node() {
    var _context;

    var _this;

    var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
        id = _ref.id,
        bind = _ref.bind,
        options = _ref.options,
        config = _ref.config;

    _classCallCheck(this, Node);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Node).call(this));
    id = id || _generateNodeId();
    options = options || {};
    Object.defineProperty(options, '_id', {
      value: id,
      writable: false,
      configurable: true,
      enumerable: true
    });
    config = config || {};
    config.logger = defaultLogger;
    _this.logger = config.logger || defaultLogger; // ** default metric is disabled

    var metric = new _metric["default"]({
      id: id
    });
    var _scope = {
      id: id,
      bind: bind,
      options: options,
      config: config,
      metric: metric,
      nodeServer: null,
      nodeClients: new Map(),
      nodeClientsAddressIndex: new Map(),
      tickWatcherMap: new Map(),
      requestWatcherMap: new Map()
    };

    _private.set(_assertThisInitialized(_this), _scope);

    (_context = _assertThisInitialized(_this), _initNodeServer).call(_context);
    return _this;
  }

  _createClass(Node, [{
    key: "getId",
    value: function getId() {
      var _private$get = _private.get(this),
          id = _private$get.id;

      return id;
    }
  }, {
    key: "getAddress",
    value: function getAddress() {
      var _private$get2 = _private.get(this),
          nodeServer = _private$get2.nodeServer;

      return nodeServer ? nodeServer.getAddress() : null;
    }
  }, {
    key: "getOptions",
    value: function getOptions() {
      var _private$get3 = _private.get(this),
          options = _private$get3.options;

      return options;
    }
  }, {
    key: "getServerInfo",
    value: function getServerInfo(_ref2) {
      var address = _ref2.address,
          id = _ref2.id;

      var _private$get4 = _private.get(this),
          nodeClients = _private$get4.nodeClients,
          nodeClientsAddressIndex = _private$get4.nodeClientsAddressIndex;

      if (!id) {
        var addressHash = (0, _md["default"])(address);
        if (!nodeClientsAddressIndex.has(addressHash)) return null;
        id = nodeClientsAddressIndex.get(addressHash);
      }

      var client = nodeClients.get(id);
      if (!client) return null;
      var serverActor = client.getServerActor();
      return serverActor ? serverActor.toJSON() : null;
    }
  }, {
    key: "getClientInfo",
    value: function getClientInfo(_ref3) {
      var id = _ref3.id;

      var _private$get5 = _private.get(this),
          nodeServer = _private$get5.nodeServer;

      var client = nodeServer.getClientById(id);
      return client ? client.toJSON() : null;
    }
  }, {
    key: "getFilteredNodes",
    value: function getFilteredNodes() {
      var _ref4 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          options = _ref4.options,
          predicate = _ref4.predicate,
          _ref4$up = _ref4.up,
          up = _ref4$up === void 0 ? true : _ref4$up,
          _ref4$down = _ref4.down,
          down = _ref4$down === void 0 ? true : _ref4$down;

      var _scope = _private.get(this);

      var nodes = new Set(); // ** if the predicate is provided we'll use it, if not then filtering will hapen based on options
      // ** options predicate is built via NodeUtils.optionsPredicateBuilder

      predicate = _underscore["default"].isFunction(predicate) ? predicate : _utils["default"].optionsPredicateBuilder(options);

      if (_scope.nodeServer && down) {
        _scope.nodeServer.getOnlineClients().forEach(function (clientNode) {
          _utils["default"].checkNodeReducer(clientNode, predicate, nodes);
        }, this);
      }

      if (_scope.nodeClients.size && up) {
        _scope.nodeClients.forEach(function (client) {
          var actorModel = client.getServerActor();

          if (actorModel && actorModel.isOnline()) {
            _utils["default"].checkNodeReducer(actorModel, predicate, nodes);
          }
        }, this);
      }

      return Array.from(nodes);
    }
  }, {
    key: "setAddress",
    value: function setAddress(bind) {
      var _private$get6 = _private.get(this),
          nodeServer = _private$get6.nodeServer;

      nodeServer ? nodeServer.setAddress(bind) : this.logger.info('No server available');
    } // ** returns promise

  }, {
    key: "bind",
    value: function bind(address) {
      var _private$get7 = _private.get(this),
          nodeServer = _private$get7.nodeServer;

      return nodeServer.bind(address);
    } // ** returns promise

  }, {
    key: "unbind",
    value: function unbind() {
      var _private$get8 = _private.get(this),
          nodeServer = _private$get8.nodeServer;

      if (!nodeServer) return _bluebird["default"].resolve();
      return nodeServer.unbind();
    } // ** connect returns the id of the connected node

  }, {
    key: "connect",
    value: function () {
      var _connect = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee() {
        var _this2 = this;

        var _ref5,
            address,
            timeout,
            reconnectionTimeout,
            _scope,
            id,
            metric,
            nodeClientsAddressIndex,
            nodeClients,
            config,
            clientConfig,
            addressHash,
            _client,
            client,
            _ref7,
            actorId,
            _args = arguments;

        return _regenerator["default"].wrap(function _callee$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _ref5 = _args.length > 0 && _args[0] !== undefined ? _args[0] : {}, address = _ref5.address, timeout = _ref5.timeout, reconnectionTimeout = _ref5.reconnectionTimeout;

                if (!(typeof address !== 'string' || address.length === 0)) {
                  _context2.next = 3;
                  break;
                }

                throw new Error("Wrong type for argument address ".concat(address));

              case 3:
                _scope = _private.get(this);
                id = _scope.id, metric = _scope.metric, nodeClientsAddressIndex = _scope.nodeClientsAddressIndex, nodeClients = _scope.nodeClients, config = _scope.config;
                clientConfig = config;
                if (reconnectionTimeout) clientConfig = Object.assign({}, config, {
                  RECONNECTION_TIMEOUT: reconnectionTimeout
                });
                address = address || 'tcp://127.0.0.1:3000';
                addressHash = (0, _md["default"])(address);

                if (!nodeClientsAddressIndex.has(addressHash)) {
                  _context2.next = 12;
                  break;
                }

                _client = nodeClients.get(nodeClientsAddressIndex.get(addressHash));
                return _context2.abrupt("return", _client.getServerActor().toJSON());

              case 12:
                client = new _client2["default"]({
                  id: id,
                  options: _scope.options,
                  config: clientConfig
                }); // ** attaching client handlers

                client.on('error', function (err) {
                  return _this2.emit('error', err);
                });
                client.on(_enum.events.SERVER_FAILURE, function (serverActor) {
                  return _this2.emit(_enum.events.SERVER_FAILURE, serverActor);
                });
                client.on(_enum.events.SERVER_STOP, function (serverActor) {
                  return _this2.emit(_enum.events.SERVER_STOP, serverActor);
                });
                client.on(_enum.events.SERVER_RECONNECT, function (serverActor) {
                  try {
                    var _addressHash = (0, _md["default"])(serverActor.address);

                    var oldId = nodeClientsAddressIndex.get(_addressHash);
                    nodeClients["delete"](oldId);
                    nodeClientsAddressIndex.set(_addressHash, serverActor.id);
                    nodeClients.set(serverActor.id, client);
                  } catch (err) {
                    _this2.logger.error('Error while handling server reconnect', err);
                  }

                  _this2.emit(_enum.events.SERVER_RECONNECT, serverActor);
                });
                client.on(_enum.events.SERVER_RECONNECT_FAILURE, function (serverActor) {
                  try {
                    nodeClients["delete"](serverActor.id);
                    nodeClientsAddressIndex["delete"]((0, _md["default"])(serverActor.address));
                  } catch (err) {
                    _this2.logger.error('Error while handling server reconnect failure', err);
                  }

                  _this2.emit(_enum.events.SERVER_RECONNECT_FAILURE, serverActor);
                });
                client.on(_enum.events.OPTIONS_SYNC, function (_ref6) {
                  var id = _ref6.id,
                      newOptions = _ref6.newOptions;
                  return _this2.emit(_enum.events.OPTIONS_SYNC, {
                    id: id,
                    newOptions: newOptions
                  });
                }); // **

                client.setMetric(metric.status);

                _addExistingListenersToClient.call(this, client);

                _context2.next = 23;
                return client.connect(address, timeout);

              case 23:
                _ref7 = _context2.sent;
                actorId = _ref7.actorId;

                _attachMetricsHandlers.call(this, client, metric);

                this.logger.info("Node connected: ".concat(this.getId(), " -> ").concat(actorId));
                nodeClientsAddressIndex.set(addressHash, actorId);
                nodeClients.set(actorId, client);
                this.emit(_enum.events.CONNECT_TO_SERVER, client.getServerActor().toJSON());
                return _context2.abrupt("return", client.getServerActor().toJSON());

              case 31:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee, this);
      }));

      function connect() {
        return _connect.apply(this, arguments);
      }

      return connect;
    }() // TODO::avar maybe disconnect from node ?

  }, {
    key: "disconnect",
    value: function () {
      var _disconnect = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee2() {
        var address,
            addressHash,
            _scope,
            nodeClientsAddressIndex,
            nodeClients,
            nodeId,
            client,
            _args2 = arguments;

        return _regenerator["default"].wrap(function _callee2$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                address = _args2.length > 0 && _args2[0] !== undefined ? _args2[0] : 'tcp://127.0.0.1:3000';

                if (!(typeof address !== 'string' || address.length === 0)) {
                  _context3.next = 3;
                  break;
                }

                throw new Error("Wrong type for argument address ".concat(address));

              case 3:
                addressHash = (0, _md["default"])(address);
                _scope = _private.get(this);
                nodeClientsAddressIndex = _scope.nodeClientsAddressIndex, nodeClients = _scope.nodeClients;

                if (nodeClientsAddressIndex.has(addressHash)) {
                  _context3.next = 8;
                  break;
                }

                return _context3.abrupt("return", true);

              case 8:
                nodeId = nodeClientsAddressIndex.get(addressHash);
                client = nodeClients.get(nodeId);
                client.removeAllListeners(_enum.events.SERVER_FAILURE);
                client.removeAllListeners(MetricType.SEND_TICK);
                client.removeAllListeners(MetricType.GOT_TICK);
                client.removeAllListeners(MetricType.SEND_REQUEST);
                client.removeAllListeners(MetricType.GOT_REQUEST);
                client.removeAllListeners(MetricType.SEND_REPLY_SUCCESS);
                client.removeAllListeners(MetricType.SEND_REPLY_ERROR);
                client.removeAllListeners(MetricType.GOT_REPLY_SUCCESS);
                client.removeAllListeners(MetricType.GOT_REPLY_ERROR);
                client.removeAllListeners(MetricType.REQUEST_TIMEOUT);
                client.removeAllListeners(MetricType.OPTIONS_SYNC);
                _context3.next = 23;
                return client.disconnect();

              case 23:
                _removeClientAllListeners.call(this, client);

                nodeClients["delete"](nodeId);
                nodeClientsAddressIndex["delete"](addressHash);
                return _context3.abrupt("return", true);

              case 27:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee2, this);
      }));

      function disconnect() {
        return _disconnect.apply(this, arguments);
      }

      return disconnect;
    }()
  }, {
    key: "stop",
    value: function () {
      var _stop = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee3() {
        var _private$get9, nodeServer, nodeClients, stopPromise;

        return _regenerator["default"].wrap(function _callee3$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _private$get9 = _private.get(this), nodeServer = _private$get9.nodeServer, nodeClients = _private$get9.nodeClients;
                stopPromise = [];
                this.disableMetrics();

                if (nodeServer.isOnline()) {
                  stopPromise.push(nodeServer.close());
                }

                nodeClients.forEach(function (client) {
                  stopPromise.push(client.close());
                }, this);
                _context4.next = 7;
                return _bluebird["default"].all(stopPromise);

              case 7:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee3, this);
      }));

      function stop() {
        return _stop.apply(this, arguments);
      }

      return stop;
    }()
  }, {
    key: "onRequest",
    value: function onRequest(requestEvent, fn) {
      var _scope = _private.get(this);

      var requestWatcherMap = _scope.requestWatcherMap,
          nodeClients = _scope.nodeClients,
          nodeServer = _scope.nodeServer;
      var requestWatcher = requestWatcherMap.get(requestEvent);

      if (!requestWatcher) {
        requestWatcher = new _sockets.Watchers(requestEvent);
        requestWatcherMap.set(requestEvent, requestWatcher);
      }

      requestWatcher.addFn(fn);
      nodeServer.onRequest(requestEvent, fn);
      nodeClients.forEach(function (client) {
        client.onRequest(requestEvent, fn);
      }, this);
    }
  }, {
    key: "offRequest",
    value: function offRequest(requestEvent, fn) {
      var _scope = _private.get(this);

      _scope.nodeServer.offRequest(requestEvent, fn);

      _scope.nodeClients.forEach(function (client) {
        client.offRequest(requestEvent, fn);
      });

      var requestWatcher = _scope.requestWatcherMap.get(requestEvent);

      if (requestWatcher) {
        requestWatcher.removeFn(fn);
      }
    }
  }, {
    key: "onTick",
    value: function onTick(event, fn) {
      var _scope = _private.get(this);

      var tickWatcherMap = _scope.tickWatcherMap,
          nodeClients = _scope.nodeClients,
          nodeServer = _scope.nodeServer;
      var tickWatcher = tickWatcherMap.get(event);

      if (!tickWatcher) {
        tickWatcher = new _sockets.Watchers(event);
        tickWatcherMap.set(event, tickWatcher);
      }

      tickWatcher.addFn(fn); // ** _scope.nodeServer is constructed in Node constructor

      nodeServer.onTick(event, fn);
      nodeClients.forEach(function (client) {
        client.onTick(event, fn);
      });
    }
  }, {
    key: "offTick",
    value: function offTick(event, fn) {
      var _scope = _private.get(this);

      _scope.nodeServer.offTick(event);

      _scope.nodeClients.forEach(function (client) {
        client.offTick(event, fn);
      }, this);

      var tickWatcher = _scope.tickWatcherMap.get(event);

      if (tickWatcher) {
        tickWatcher.removeFn(fn);
      }
    }
  }, {
    key: "request",
    value: function () {
      var _request = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee4() {
        var _ref8,
            to,
            event,
            data,
            timeout,
            _scope,
            nodeServer,
            nodeClients,
            clientActor,
            _args4 = arguments;

        return _regenerator["default"].wrap(function _callee4$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _ref8 = _args4.length > 0 && _args4[0] !== undefined ? _args4[0] : {}, to = _ref8.to, event = _ref8.event, data = _ref8.data, timeout = _ref8.timeout;
                _scope = _private.get(this);
                nodeServer = _scope.nodeServer, nodeClients = _scope.nodeClients;
                clientActor = _getClientByNode.call(this, to);

                if (!clientActor) {
                  _context5.next = 6;
                  break;
                }

                return _context5.abrupt("return", nodeServer.request({
                  to: clientActor.getId(),
                  event: event,
                  data: data,
                  timeout: timeout
                }));

              case 6:
                if (!nodeClients.has(to)) {
                  _context5.next = 8;
                  break;
                }

                return _context5.abrupt("return", nodeClients.get(to).request({
                  event: event,
                  data: data,
                  timeout: timeout
                }));

              case 8:
                throw new _errors.ZeronodeError({
                  message: "Node with id '".concat(to, "' is not found."),
                  code: _errors.ErrorCodes.NODE_NOT_FOUND
                });

              case 9:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee4, this);
      }));

      function request() {
        return _request.apply(this, arguments);
      }

      return request;
    }()
  }, {
    key: "tick",
    value: function tick() {
      var _ref9 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          to = _ref9.to,
          event = _ref9.event,
          data = _ref9.data;

      var _scope = _private.get(this);

      var nodeServer = _scope.nodeServer,
          nodeClients = _scope.nodeClients;

      var clientActor = _getClientByNode.call(this, to);

      if (clientActor) {
        return nodeServer.tick({
          to: clientActor.getId(),
          event: event,
          data: data
        });
      }

      if (nodeClients.has(to)) {
        return nodeClients.get(to).tick({
          event: event,
          data: data
        });
      }

      throw new _errors.ZeronodeError({
        message: "Node with id '".concat(to, "' is not found."),
        code: _errors.ErrorCodes.NODE_NOT_FOUND
      });
    }
  }, {
    key: "requestAny",
    value: function () {
      var _requestAny = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee5() {
        var _ref10,
            event,
            data,
            timeout,
            filter,
            _ref10$down,
            down,
            _ref10$up,
            up,
            nodesFilter,
            filteredNodes,
            to,
            _args5 = arguments;

        return _regenerator["default"].wrap(function _callee5$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                _ref10 = _args5.length > 0 && _args5[0] !== undefined ? _args5[0] : {}, event = _ref10.event, data = _ref10.data, timeout = _ref10.timeout, filter = _ref10.filter, _ref10$down = _ref10.down, down = _ref10$down === void 0 ? true : _ref10$down, _ref10$up = _ref10.up, up = _ref10$up === void 0 ? true : _ref10$up;
                nodesFilter = {
                  down: down,
                  up: up
                };

                if (_underscore["default"].isFunction(filter)) {
                  nodesFilter.predicate = filter;
                } else {
                  nodesFilter.options = filter || {};
                }

                filteredNodes = this.getFilteredNodes(nodesFilter);

                if (filteredNodes.length) {
                  _context6.next = 6;
                  break;
                }

                throw new _errors.ZeronodeError({
                  message: "Node with filter is not found.",
                  code: _errors.ErrorCodes.NODE_NOT_FOUND
                });

              case 6:
                // ** find the node id where the request will be sent
                to = _getWinnerNode.call(this, filteredNodes, event);
                return _context6.abrupt("return", this.request({
                  to: to,
                  event: event,
                  data: data,
                  timeout: timeout
                }));

              case 8:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee5, this);
      }));

      function requestAny() {
        return _requestAny.apply(this, arguments);
      }

      return requestAny;
    }()
  }, {
    key: "requestDownAny",
    value: function () {
      var _requestDownAny = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee6() {
        var _ref11,
            event,
            data,
            timeout,
            filter,
            result,
            _args6 = arguments;

        return _regenerator["default"].wrap(function _callee6$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _ref11 = _args6.length > 0 && _args6[0] !== undefined ? _args6[0] : {}, event = _ref11.event, data = _ref11.data, timeout = _ref11.timeout, filter = _ref11.filter;
                _context7.next = 3;
                return this.requestAny({
                  event: event,
                  data: data,
                  timeout: timeout,
                  filter: filter,
                  down: true,
                  up: false
                });

              case 3:
                result = _context7.sent;
                return _context7.abrupt("return", result);

              case 5:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee6, this);
      }));

      function requestDownAny() {
        return _requestDownAny.apply(this, arguments);
      }

      return requestDownAny;
    }()
  }, {
    key: "requestUpAny",
    value: function () {
      var _requestUpAny = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee7() {
        var _ref12,
            event,
            data,
            timeout,
            filter,
            result,
            _args7 = arguments;

        return _regenerator["default"].wrap(function _callee7$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                _ref12 = _args7.length > 0 && _args7[0] !== undefined ? _args7[0] : {}, event = _ref12.event, data = _ref12.data, timeout = _ref12.timeout, filter = _ref12.filter;
                _context8.next = 3;
                return this.requestAny({
                  event: event,
                  data: data,
                  timeout: timeout,
                  filter: filter,
                  down: false,
                  up: true
                });

              case 3:
                result = _context8.sent;
                return _context8.abrupt("return", result);

              case 5:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee7, this);
      }));

      function requestUpAny() {
        return _requestUpAny.apply(this, arguments);
      }

      return requestUpAny;
    }()
  }, {
    key: "tickAny",
    value: function tickAny() {
      var _ref13 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          event = _ref13.event,
          data = _ref13.data,
          filter = _ref13.filter,
          _ref13$down = _ref13.down,
          down = _ref13$down === void 0 ? true : _ref13$down,
          _ref13$up = _ref13.up,
          up = _ref13$up === void 0 ? true : _ref13$up;

      var nodesFilter = {
        down: down,
        up: up
      };

      if (_underscore["default"].isFunction(filter)) {
        nodesFilter.predicate = filter;
      } else {
        nodesFilter.options = filter || {};
      }

      var filteredNodes = this.getFilteredNodes(nodesFilter);

      if (!filteredNodes.length) {
        throw new _errors.ZeronodeError({
          message: "Node with filter is not found.",
          code: _errors.ErrorCodes.NODE_NOT_FOUND
        });
      }

      var nodeId = _getWinnerNode.call(this, filteredNodes, event);

      return this.tick({
        to: nodeId,
        event: event,
        data: data
      });
    }
  }, {
    key: "tickDownAny",
    value: function tickDownAny() {
      var _ref14 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          event = _ref14.event,
          data = _ref14.data,
          filter = _ref14.filter;

      return this.tickAny({
        event: event,
        data: data,
        filter: filter,
        down: true,
        up: false
      });
    }
  }, {
    key: "tickUpAny",
    value: function tickUpAny() {
      var _ref15 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          event = _ref15.event,
          data = _ref15.data,
          filter = _ref15.filter;

      return this.tickAny({
        event: event,
        data: data,
        filter: filter,
        down: false,
        up: true
      });
    }
  }, {
    key: "tickAll",
    value: function tickAll() {
      var _this3 = this;

      var _ref16 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          event = _ref16.event,
          data = _ref16.data,
          filter = _ref16.filter,
          _ref16$down = _ref16.down,
          down = _ref16$down === void 0 ? true : _ref16$down,
          _ref16$up = _ref16.up,
          up = _ref16$up === void 0 ? true : _ref16$up;

      var nodesFilter = {
        down: down,
        up: up
      };

      if (_underscore["default"].isFunction(filter)) {
        nodesFilter.predicate = filter;
      } else {
        nodesFilter.options = filter || {};
      }

      var filteredNodes = this.getFilteredNodes(nodesFilter);
      var tickPromises = [];
      filteredNodes.forEach(function (nodeId) {
        tickPromises.push(_this3.tick({
          to: nodeId,
          event: event,
          data: data
        }));
      }, this);
      return _bluebird["default"].all(tickPromises);
    }
  }, {
    key: "tickDownAll",
    value: function tickDownAll() {
      var _ref17 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          event = _ref17.event,
          data = _ref17.data,
          filter = _ref17.filter;

      return this.tickAll({
        event: event,
        data: data,
        filter: filter,
        down: true,
        up: false
      });
    }
  }, {
    key: "tickUpAll",
    value: function tickUpAll() {
      var _ref18 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          event = _ref18.event,
          data = _ref18.data,
          filter = _ref18.filter;

      return this.tickAll({
        event: event,
        data: data,
        filter: filter,
        down: false,
        up: true
      });
    }
  }, {
    key: "enableMetrics",
    value: function enableMetrics(flushInterval) {
      var _scope = _private.get(this);

      var metric = _scope.metric,
          nodeClients = _scope.nodeClients,
          nodeServer = _scope.nodeServer;
      metric.enable(flushInterval);
      nodeClients.forEach(function (client) {
        client.setMetric(true);
      }, this);
      nodeServer.setMetric(true);
    }
  }, {
    key: "disableMetrics",
    value: function disableMetrics() {
      var _private$get10 = _private.get(this),
          metric = _private$get10.metric,
          nodeClients = _private$get10.nodeClients,
          nodeServer = _private$get10.nodeServer;

      metric.disable();
      nodeClients.forEach(function (client) {
        client.setMetric(false);
      }, this);
      nodeServer.setMetric(false);
    }
  }, {
    key: "setOptions",
    value: function () {
      var _setOptions = _asyncToGenerator(
      /*#__PURE__*/
      _regenerator["default"].mark(function _callee8() {
        var options,
            _scope,
            nodeServer,
            nodeClients,
            _args8 = arguments;

        return _regenerator["default"].wrap(function _callee8$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                options = _args8.length > 0 && _args8[0] !== undefined ? _args8[0] : {};
                _scope = _private.get(this);
                _scope.options = options;
                Object.defineProperty(options, '_id', {
                  value: _scope.id,
                  writable: false,
                  configurable: true,
                  enumerable: true
                });
                nodeServer = _scope.nodeServer, nodeClients = _scope.nodeClients;
                nodeServer.setOptions(options);
                nodeClients.forEach(function (client) {
                  client.setOptions(options);
                }, this);

              case 7:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee8, this);
      }));

      function setOptions() {
        return _setOptions.apply(this, arguments);
      }

      return setOptions;
    }()
  }, {
    key: "metric",
    get: function get() {
      var _private$get11 = _private.get(this),
          metric = _private$get11.metric;

      return metric;
    }
  }]);

  return Node;
}(_events.EventEmitter); // ** PRIVATE FUNCTIONS


exports["default"] = Node;

function _initNodeServer() {
  var _this4 = this;

  var _scope = _private.get(this);

  var id = _scope.id,
      bind = _scope.bind,
      options = _scope.options,
      metric = _scope.metric,
      config = _scope.config;
  var nodeServer = new _server["default"]({
    id: id,
    bind: bind,
    options: options,
    config: config
  }); // ** handlers for nodeServer

  nodeServer.on('error', function (err) {
    return _this4.emit('error', err);
  });
  nodeServer.on(_enum.events.CLIENT_FAILURE, function (clientActor) {
    return _this4.emit(_enum.events.CLIENT_FAILURE, clientActor);
  });
  nodeServer.on(_enum.events.CLIENT_CONNECTED, function (clientActor) {
    return _this4.emit(_enum.events.CLIENT_CONNECTED, clientActor);
  });
  nodeServer.on(_enum.events.CLIENT_STOP, function (clientActor) {
    return _this4.emit(_enum.events.CLIENT_STOP, clientActor);
  });
  nodeServer.on(_enum.events.OPTIONS_SYNC, function (_ref19) {
    var id = _ref19.id,
        newOptions = _ref19.newOptions;
    return _this4.emit(_enum.events.OPTIONS_SYNC, {
      id: id,
      newOptions: newOptions
    });
  }); // ** enabling metrics

  nodeServer.setMetric(metric.status);

  _attachMetricsHandlers.call(this, nodeServer, metric);

  _scope.nodeServer = nodeServer;
}

function _getClientByNode(nodeId) {
  var _scope = _private.get(this);

  var actors = _scope.nodeServer.getOnlineClients().filter(function (actor) {
    var node = actor.getId();
    return node === nodeId;
  });

  if (!actors.length) {
    return null;
  }

  if (actors.length > 1) {
    return this.logger.warn("We should have just 1 client from 1 node");
  }

  return actors[0];
}

function _generateNodeId() {
  return _animalId["default"].getId();
} // TODO::avar optimize this


function _getWinnerNode(nodeIds, tag) {
  var len = nodeIds.length;
  var idx = Math.floor(Math.random() * len);
  return nodeIds[idx];
}

function _addExistingListenersToClient(client) {
  var _this5 = this;

  var _scope = _private.get(this); // ** adding previously added onTick-s for this client to


  _scope.tickWatcherMap.forEach(function (tickWatcher, event) {
    // ** TODO what about order of functions ?
    tickWatcher.getFnMap().forEach(function (index, fn) {
      client.onTick(event, fn.bind(_this5));
    }, _this5);
  }, this); // ** adding previously added onRequests-s for this client to


  _scope.requestWatcherMap.forEach(function (requestWatcher, requestEvent) {
    // ** TODO what about order of functions ?
    requestWatcher.getFnMap().forEach(function (index, fn) {
      client.onRequest(requestEvent, fn.bind(_this5));
    }, _this5);
  }, this);
}

function _removeClientAllListeners(client) {
  var _scope = _private.get(this); // ** removing all handlers


  _scope.tickWatcherMap.forEach(function (tickWatcher, event) {
    client.offTick(event);
  }, this); // ** removing all handlers


  _scope.requestWatcherMap.forEach(function (requestWatcher, requestEvent) {
    client.offRequest(requestEvent);
  }, this);
}

function _attachMetricsHandlers(socket, metric) {
  var _this6 = this;

  socket.on(MetricType.SEND_TICK, function (envelop) {
    _this6.emit(MetricType.SEND_TICK, envelop);

    metric.sendTick(envelop);
  });
  socket.on(MetricType.SEND_REQUEST, function (envelop) {
    _this6.emit(MetricType.SEND_REQUEST, envelop);

    metric.sendRequest(envelop);
  });
  socket.on(MetricType.SEND_REPLY_SUCCESS, function (envelop) {
    _this6.emit(MetricType.SEND_REPLY_SUCCESS, envelop);

    metric.sendReplySuccess(envelop);
  });
  socket.on(MetricType.SEND_REPLY_ERROR, function (envelop) {
    _this6.emit(MetricType.SEND_REPLY_ERROR, envelop);

    metric.sendReplyError(envelop);
  });
  socket.on(MetricType.REQUEST_TIMEOUT, function (envelop) {
    _this6.emit(MetricType.REQUEST_TIMEOUT, envelop);

    metric.requestTimeout(envelop);
  });
  socket.on(MetricType.GOT_TICK, function (envelop) {
    _this6.emit(MetricType.GOT_TICK, envelop);

    metric.gotTick(envelop);
  });
  socket.on(MetricType.GOT_REQUEST, function (envelop) {
    _this6.emit(MetricType.GOT_REQUEST, envelop);

    metric.gotRequest(envelop);
  });
  socket.on(MetricType.GOT_REPLY_SUCCESS, function (envelop) {
    _this6.emit(MetricType.GOT_REPLY_SUCCESS, envelop);

    metric.gotReplySuccess(envelop);
  });
  socket.on(MetricType.GOT_REPLY_ERROR, function (envelop) {
    _this6.emit(MetricType.GOT_REPLY_ERROR, envelop);

    metric.gotReplyError(envelop);
  });
}