"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "SocketEvent", {
  enumerable: true,
  get: function get() {
    return _events["default"];
  }
});
exports["default"] = exports.Socket = void 0;

var _underscore = _interopRequireDefault(require("underscore"));

var _animalId = _interopRequireDefault(require("animal-id"));

var _patternEmitter = _interopRequireDefault(require("pattern-emitter"));

var _errors = require("../errors");

var _events = _interopRequireDefault(require("./events"));

var _envelope = _interopRequireDefault(require("./envelope"));

var _enum = require("./enum");

var _watchers = _interopRequireDefault(require("./watchers"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var _private = new WeakMap();

function _calculateLatency(_ref) {
  var sendTime = _ref.sendTime,
      getTime = _ref.getTime,
      replyTime = _ref.replyTime,
      replyGetTime = _ref.replyGetTime;
  var processTime = replyTime[0] * 10e9 + replyTime[1] - (getTime[0] * 10e9 + getTime[1]);
  var requestTime = replyGetTime[0] * 10e9 + replyGetTime[1] - (sendTime[0] * 10e9 + sendTime[1]);
  return {
    process: processTime,
    latency: requestTime - processTime
  };
}

var nop = function nop() {};
/**
 *
 * @param envelop: Object
 * @param type: Enum(-1 = timeout, 0 = send, 1 = got)
 */


function emitMetric(envelop) {
  var type = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  var event = '';
  if (envelop.mainEvent) return;

  switch (envelop.type) {
    case _enum.EnvelopType.TICK:
      event = !type ? _enum.MetricType.SEND_TICK : _enum.MetricType.GOT_TICK;
      break;

    case _enum.EnvelopType.REQUEST:
      if (type === -1) {
        event = _enum.MetricType.REQUEST_TIMEOUT;
        break;
      }

      event = !type ? _enum.MetricType.SEND_REQUEST : _enum.MetricType.GOT_REQUEST;
      break;

    case _enum.EnvelopType.RESPONSE:
      event = !type ? _enum.MetricType.SEND_REPLY_SUCCESS : _enum.MetricType.GOT_REPLY_SUCCESS;
      break;

    case _enum.EnvelopType.ERROR:
      event = !type ? _enum.MetricType.SEND_REPLY_ERROR : _enum.MetricType.GOT_REPLY_ERROR;
  }

  this.emit(event, envelop);
}

function buildSocketEventHandler(eventName) {
  var _this = this;

  var handler = function handler(fd, endpoint) {
    if (_this.debugMode()) {
      _this.logger.info("Emitted '".concat(eventName, "' on socket '").concat(_this.getId(), "'"));
    }

    _this.emit(eventName, {
      fd: fd,
      endpoint: endpoint
    });
  };

  return handler.bind(this);
}

var Socket =
/*#__PURE__*/
function (_EventEmitter) {
  _inherits(Socket, _EventEmitter);

  _createClass(Socket, null, [{
    key: "generateSocketId",
    value: function generateSocketId() {
      return _animalId["default"].getId();
    }
  }]);

  function Socket() {
    var _context;

    var _this2;

    var _ref2 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
        id = _ref2.id,
        socket = _ref2.socket,
        config = _ref2.config,
        options = _ref2.options;

    _classCallCheck(this, Socket);

    _this2 = _possibleConstructorReturn(this, _getPrototypeOf(Socket).call(this));
    options = options || {};
    config = config || {}; // ** creating the socket

    var socketId = id || Socket.generateSocketId();
    socket.identity = socketId;
    socket.on('message', (_context = _assertThisInitialized(_this2), onSocketMessage).bind(_context));
    var _scope = {
      id: socketId,
      socket: socket,
      config: config,
      options: options,
      logger: null,
      online: false,
      metric: nop,
      isDebugMode: false,
      monitorRestartInterval: null,
      requests: new Map(),
      requestWatcherMap: {
        main: new Map(),
        custom: new Map()
      },
      tickEmitter: {
        main: new _patternEmitter["default"](),
        custom: new _patternEmitter["default"]()
      }
    };

    _private.set(_assertThisInitialized(_this2), _scope); // ** setting the logger as soon as possible


    _this2.setLogger(config.logger);

    _this2.debugMode(false);

    return _this2;
  }

  _createClass(Socket, [{
    key: "getId",
    value: function getId() {
      var _private$get = _private.get(this),
          id = _private$get.id;

      return id;
    }
  }, {
    key: "setOnline",
    value: function setOnline() {
      var _scope = _private.get(this);

      _scope.online = Date.now();
    }
  }, {
    key: "setOffline",
    value: function setOffline() {
      var _scope = _private.get(this);

      _scope.online = false;
    }
  }, {
    key: "isOnline",
    value: function isOnline() {
      var _private$get2 = _private.get(this),
          online = _private$get2.online;

      return !!online;
    }
  }, {
    key: "setOptions",
    value: function setOptions() {
      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

      var _scope = _private.get(this);

      _scope.options = options;
    }
  }, {
    key: "getOptions",
    value: function getOptions() {
      var _private$get3 = _private.get(this),
          options = _private$get3.options;

      return options;
    }
  }, {
    key: "getConfig",
    value: function getConfig() {
      var _private$get4 = _private.get(this),
          config = _private$get4.config;

      return config;
    }
  }, {
    key: "setMetric",
    value: function setMetric(status) {
      var _scope = _private.get(this);

      _scope.metric = status ? emitMetric.bind(this) : nop;
    }
  }, {
    key: "setLogger",
    value: function setLogger(logger) {
      this.logger = logger || console;
    }
  }, {
    key: "debugMode",
    value: function debugMode(val) {
      var _scope = _private.get(this);

      if (val) {
        _scope.isDebugMode = !!val;
      } else {
        return _scope.isDebugMode;
      }
    }
  }, {
    key: "request",
    value: function request(envelop, reqTimeout) {
      var _this3 = this;

      var _private$get5 = _private.get(this),
          id = _private$get5.id,
          requests = _private$get5.requests,
          metric = _private$get5.metric,
          config = _private$get5.config;

      reqTimeout = reqTimeout || config.REQUEST_TIMEOUT || _enum.Timeouts.REQUEST_TIMEOUT;

      if (!this.isOnline()) {
        var err = new Error("Sending failed as socket '".concat(this.getId(), "' is not online"));
        return Promise.reject(new _errors.ZeronodeError({
          socketId: id,
          error: err,
          code: _errors.ErrorCodes.SOCKET_ISNOT_ONLINE
        }));
      }

      var envelopId = envelop.getId();
      return new Promise(function (resolve, reject) {
        var timeout = setTimeout(function () {
          if (requests.has(envelopId)) {
            var requestObj = requests.get(envelopId);
            requests["delete"](envelopId);
            metric(envelop.toJSON(), -1);
            var requestTimeoutedError = new Error("Request envelop '".concat(envelopId, "' timeouted on socket '").concat(_this3.getId(), "'"));
            requestObj.reject(new _errors.ZeronodeError({
              socketId: _this3.getId(),
              envelopId: envelopId,
              error: requestTimeoutedError,
              code: _errors.ErrorCodes.REQUEST_TIMEOUTED
            }));
          }
        }, reqTimeout);
        requests.set(envelopId, {
          resolve: resolve,
          reject: reject,
          timeout: timeout,
          sendTime: process.hrtime()
        });

        _this3.sendEnvelop(envelop);
      });
    }
  }, {
    key: "tick",
    value: function tick(envelop) {
      var socketId = this.getId();

      if (!this.isOnline()) {
        var socketNotOnlineError = new Error("Sending failed as socket ".concat(socketId, " is not online"));
        throw new _errors.ZeronodeError({
          socketId: socketId,
          error: socketNotOnlineError,
          code: _errors.ErrorCodes.SOCKET_ISNOT_ONLINE
        });
      }

      this.sendEnvelop(envelop);
    }
  }, {
    key: "sendEnvelop",
    value: function sendEnvelop(envelop) {
      var _private$get6 = _private.get(this),
          socket = _private$get6.socket,
          metric = _private$get6.metric;

      var msg = this.getSocketMsg(envelop);
      var envelopJSON = envelop.toJSON();

      if (msg instanceof Buffer) {
        envelopJSON.size = msg.length;
      } else {
        envelopJSON.size = msg[2].length;
      }

      metric(envelopJSON);
      socket.send(msg);
    }
  }, {
    key: "attachSocketMonitor",
    value: function attachSocketMonitor() {
      var _this4 = this;

      var _scope = _private.get(this);

      var config = _scope.config,
          socket = _scope.socket; // ** start monitoring socket events

      var monitorTimeout = config.MONITOR_TIMEOUT || _enum.Timeouts.MONITOR_TIMEOUT;
      var monitorRestartTimeout = config.MONITOR_RESTART_TIMEOUT || _enum.Timeouts.MONITOR_RESTART_TIMEOUT; // ** start socket monitoring

      socket.monitor(monitorTimeout, 0); // ** Handle monitor error and restart it

      socket.on('monitor_error', function () {
        _this4.logger.warn("Restarting monitor after ".concat(monitorRestartTimeout, " on socket ").concat(_this4.getId()));

        _scope.monitorRestartInterval = setTimeout(function () {
          return socket.monitor(monitorTimeout, 0);
        }, monitorRestartTimeout);
      });
      socket.on('connect', buildSocketEventHandler.call(this, _events["default"].CONNECT));
      socket.on('disconnect', buildSocketEventHandler.call(this, _events["default"].DISCONNECT));
      socket.on('connect_delay', buildSocketEventHandler.call(this, _events["default"].CONNECT_DELAY));
      socket.on('connect_retry', buildSocketEventHandler.call(this, _events["default"].CONNECT_RETRY));
      socket.on('listen', buildSocketEventHandler.call(this, _events["default"].LISTEN));
      socket.on('bind_error', buildSocketEventHandler.call(this, _events["default"].BIND_ERROR));
      socket.on('accept', buildSocketEventHandler.call(this, _events["default"].ACCEPT));
      socket.on('accept_error', buildSocketEventHandler.call(this, _events["default"].ACCEPT_ERROR));
      socket.on('close', buildSocketEventHandler.call(this, _events["default"].CLOSE));
      socket.on('close_error', buildSocketEventHandler.call(this, _events["default"].CLOSE_ERROR));
    }
  }, {
    key: "detachSocketMonitor",
    value: function detachSocketMonitor() {
      var _private$get7 = _private.get(this),
          socket = _private$get7.socket,
          monitorRestartInterval = _private$get7.monitorRestartInterval; // ** remove all listeners


      socket.removeAllListeners('connect');
      socket.removeAllListeners('disconnect');
      socket.removeAllListeners('connect_delay');
      socket.removeAllListeners('connect_retry');
      socket.removeAllListeners('listen');
      socket.removeAllListeners('bind_error');
      socket.removeAllListeners('accept');
      socket.removeAllListeners('accept_error');
      socket.removeAllListeners('close');
      socket.removeAllListeners('close_error'); // ** if during closing there is a monitor restart scheduled then clear the schedule

      if (monitorRestartInterval) clearInterval(monitorRestartInterval);
      socket.unmonitor();
    }
  }, {
    key: "close",
    value: function close() {
      this.detachSocketMonitor();
    }
  }, {
    key: "onRequest",
    value: function onRequest(endpoint, fn) {
      var main = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

      // ** function will called with argument  request = {body, reply}
      if (!(endpoint instanceof RegExp)) {
        endpoint = endpoint.toString();
      }

      var _private$get8 = _private.get(this),
          requestWatcherMap = _private$get8.requestWatcherMap;

      var watcherMap = main ? requestWatcherMap.main : requestWatcherMap.custom;
      var requestWatcher = watcherMap.get(endpoint);

      if (!requestWatcher) {
        requestWatcher = new _watchers["default"](endpoint);
        watcherMap.set(endpoint, requestWatcher);
      }

      requestWatcher.addFn(fn);
    }
  }, {
    key: "offRequest",
    value: function offRequest(endpoint, fn) {
      var main = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

      var _private$get9 = _private.get(this),
          requestWatcherMap = _private$get9.requestWatcherMap;

      var watcherMap = main ? requestWatcherMap.main : requestWatcherMap.custom;

      if (_underscore["default"].isFunction(fn)) {
        var endpointWatcher = watcherMap.get(endpoint);
        if (!endpointWatcher) return;
        endpointWatcher.removeFn(fn);
        return;
      }

      watcherMap["delete"](endpoint);
    }
  }, {
    key: "onTick",
    value: function onTick(event, fn) {
      var main = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

      var _private$get10 = _private.get(this),
          tickEmitter = _private$get10.tickEmitter;

      main ? tickEmitter.main.on(event, fn) : tickEmitter.custom.on(event, fn);
    }
  }, {
    key: "offTick",
    value: function offTick(event, fn) {
      var main = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

      var _private$get11 = _private.get(this),
          tickEmitter = _private$get11.tickEmitter;

      var eventTickEmitter = main ? tickEmitter.main : tickEmitter.custom;

      if (_underscore["default"].isFunction(fn)) {
        eventTickEmitter.removeListener(event, fn);
        return;
      }

      eventTickEmitter.removeAllListeners(event);
    }
  }]);

  return Socket;
}(_patternEmitter["default"]); //* * Handlers of specific envelop msg-es
//* * when socket is dealer identity is empty
//* * when socket is router, identity is the dealer which sends data


exports.Socket = Socket;

function onSocketMessage(empty, envelopBuffer) {
  var _private$get12 = _private.get(this),
      metric = _private$get12.metric,
      tickEmitter = _private$get12.tickEmitter;

  var _Envelop$readMetaFrom = _envelope["default"].readMetaFromBuffer(envelopBuffer),
      type = _Envelop$readMetaFrom.type,
      id = _Envelop$readMetaFrom.id,
      owner = _Envelop$readMetaFrom.owner,
      recipient = _Envelop$readMetaFrom.recipient,
      tag = _Envelop$readMetaFrom.tag,
      mainEvent = _Envelop$readMetaFrom.mainEvent;

  var envelop = new _envelope["default"]({
    type: type,
    id: id,
    owner: owner,
    recipient: recipient,
    tag: tag,
    mainEvent: mainEvent
  });

  var envelopData = _envelope["default"].readDataFromBuffer(envelopBuffer);

  envelop.setData(envelopData);
  var envelopJSON = envelop.toJSON();
  envelopJSON.size = envelopBuffer.length;

  switch (type) {
    case _enum.EnvelopType.TICK:
      metric(envelopJSON, 1);

      if (mainEvent) {
        tickEmitter.main.emit(tag, envelopData);
      } else {
        tickEmitter.custom.emit(tag, envelopData, {
          id: owner,
          event: tag
        });
      }

      break;

    case _enum.EnvelopType.REQUEST:
      metric(envelopJSON, 1); // ** if metric is enabled then emit it

      syncEnvelopHandler.call(this, envelop);
      break;

    case _enum.EnvelopType.RESPONSE:
    case _enum.EnvelopType.ERROR:
      envelop.size = envelopBuffer.length;
      responseEnvelopHandler.call(this, envelop);
      break;
  }
}

function syncEnvelopHandler(envelop) {
  var self = this;
  var getTime = process.hrtime();
  var prevOwner = envelop.getOwner();
  var handlers = determineHandlersByTag.call(self, envelop.getTag(), envelop.isMain());
  if (!handlers.length) return;
  var requestOb = {
    head: {
      id: envelop.getOwner(),
      event: envelop.getTag()
    },
    body: envelop.getData(),
    reply: function reply(response) {
      envelop.setRecipient(prevOwner);
      envelop.setOwner(self.getId());
      envelop.setType(_enum.EnvelopType.RESPONSE);
      envelop.setData({
        getTime: getTime,
        replyTime: process.hrtime(),
        data: response
      });
      self.sendEnvelop(envelop);
    },
    error: function error(err) {
      envelop.setRecipient(prevOwner);
      envelop.setOwner(self.getId());
      envelop.setType(_enum.EnvelopType.ERROR);
      envelop.setData({
        getTime: getTime,
        replyTime: process.hrtime(),
        data: err
      });
      self.sendEnvelop(envelop);
    },
    next: function next(err) {
      if (err) {
        return requestOb.error(err);
      }

      if (!handlers.length) {
        var noHandlerErr = new Error("There is no handlers available as to process next() on socket '".concat(self.getId(), "'"));
        throw new _errors.ZeronodeError({
          socketId: self.getId(),
          code: _errors.ErrorCodes.NO_NEXT_HANDLER_AVAILABLE,
          error: noHandlerErr
        });
      }

      handlers.shift()(requestOb);
    }
  };
  handlers.shift()(requestOb);
}

function determineHandlersByTag(tag) {
  var main = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
  var handlers = [];

  var _private$get13 = _private.get(this),
      requestWatcherMap = _private$get13.requestWatcherMap;

  var watcherMap = main ? requestWatcherMap.main : requestWatcherMap.custom;
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = watcherMap.keys()[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var endpoint = _step.value;

      if (endpoint instanceof RegExp) {
        if (endpoint.test(tag)) {
          watcherMap.get(endpoint).getFnMap().forEach(function (index, fnKey) {
            handlers.push({
              index: index,
              fnKey: fnKey
            });
          });
        }
      } else if (endpoint === tag) {
        watcherMap.get(endpoint).getFnMap().forEach(function (index, fnKey) {
          handlers.push({
            index: index,
            fnKey: fnKey
          });
        });
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator["return"] != null) {
        _iterator["return"]();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return handlers.sort(function (a, b) {
    return a.index - b.index;
  }).map(function (ob) {
    return ob.fnKey;
  });
}

function responseEnvelopHandler(envelop) {
  var _private$get14 = _private.get(this),
      requests = _private$get14.requests,
      metric = _private$get14.metric;

  var id = envelop.getId();

  if (!requests.has(id)) {
    // ** TODO:: metric
    return this.logger.warn("Response ".concat(id, " is probably time outed"));
  } //* * requestObj is like {resolve, reject, timeout : clearRequestTimeout}


  var _requests$get = requests.get(id),
      timeout = _requests$get.timeout,
      sendTime = _requests$get.sendTime,
      resolve = _requests$get.resolve,
      reject = _requests$get.reject; // ** getTime is the time when message arrives to server
  // ** replyTime is the time when message is send from server


  var gotReplyMetric = envelop.toJSON();
  var _gotReplyMetric$data = gotReplyMetric.data,
      getTime = _gotReplyMetric$data.getTime,
      replyTime = _gotReplyMetric$data.replyTime;

  var duration = _calculateLatency({
    sendTime: sendTime,
    getTime: getTime,
    replyTime: replyTime,
    replyGetTime: process.hrtime()
  });

  gotReplyMetric.data = {
    data: gotReplyMetric.data,
    duration: duration
  };
  gotReplyMetric.size = envelop.size;
  metric(gotReplyMetric, 1);
  clearTimeout(timeout);
  requests["delete"](id);

  var _envelop$getData = envelop.getData(),
      data = _envelop$getData.data; //* * resolving request promise with response data


  envelop.getType() === _enum.EnvelopType.ERROR ? reject(data) : resolve(data);
} // ** exports


var _default = {
  SocketEvent: _events["default"],
  Socket: Socket
};
exports["default"] = _default;