"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _underscore = _interopRequireDefault(require("underscore"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

// ** we use this to check if node is satisfying the predicate and adding node id into accumulatorSet
var checkNodeReducer = function checkNodeReducer(node, predicate, accumulatorSet) {
  var nodeOptions = node.getOptions();

  if (_underscore["default"].isFunction(predicate) && predicate(nodeOptions)) {
    accumulatorSet.add(node.getId());
  }
};

var optionsPredicateBuilder = function optionsPredicateBuilder(options) {
  return function (nodeOptions) {
    var optionsKeysArray = Object.keys(options);

    var notsatisfying = _underscore["default"].find(optionsKeysArray, function (optionKey) {
      var optionValue = options[optionKey]; // ** which could also not exist

      var nodeOptionValue = nodeOptions[optionKey];

      if (nodeOptionValue) {
        if (_underscore["default"].isRegExp(optionValue)) {
          return !optionValue.test(nodeOptionValue);
        }

        if (_underscore["default"].isString(optionValue) || _underscore["default"].isNumber(optionValue)) {
          return optionValue !== nodeOptionValue;
        }

        if (_underscore["default"].isObject(optionValue)) {
          return !!_underscore["default"].find(optionValue, function (value, operator) {
            switch (operator) {
              case '$eq':
                return value !== nodeOptionValue;

              case '$ne':
                return value === nodeOptionValue;

              case '$aeq':
                return value != nodeOptionValue;

              case '$gt':
                return value >= nodeOptionValue;

              case '$gte':
                return value > nodeOptionValue;

              case '$lt':
                return value <= nodeOptionValue;

              case '$lte':
                return value < nodeOptionValue;

              case '$between':
                return value[0] >= nodeOptionValue || value[1] <= nodeOptionValue;

              case '$regex':
                return !value.test(nodeOptionValue);

              case '$in':
                return value.indexOf(nodeOptionValue) === -1;

              case '$nin':
                return value.indexOf(nodeOptionValue) !== -1;

              case '$contains':
                return nodeOptionValue.indexOf(value) === -1;

              case '$containsAny':
                return !_underscore["default"].find(value, function (v) {
                  return nodeOptionValue.indexOf(v) !== -1;
                });

              case '$containsNone':
                return !!_underscore["default"].find(value, function (v) {
                  return nodeOptionValue.indexOf(v) !== -1;
                });
            }
          });
        }
      }

      return true;
    });

    return !notsatisfying;
  };
};

var _default = {
  checkNodeReducer: checkNodeReducer,
  optionsPredicateBuilder: optionsPredicateBuilder
};
exports["default"] = _default;